@@ -257,11 +257,34 @@ class TextPreprocessor:
           + r'{,138})\b(?!\S*' + re_char_delim + ')'
         )
         self.TAG_NAMETAG = self.register_tag('EntityNametag', mask='ссылка')
+        re_1 = r'\s*[A-ZЁА-Я][^"]+[.!?](?:\s*[A-ZЁА-Я][^"])*'
+        self.RE_QUOTATION = re_compile(r'''(?xmu)
+            (?:(")({0})("))|    # 1 - 3
+            (?:(``)({0})(''))|  # 4 - 6
+            (?:(«)({0})(»))|    # 7 - 9
+            (?:(„)({0})(“))     # 10 - 12
+        '''.format(re_1))
+        self.TAG_QUOTATION_START = self.register_tag('QuotationStart', '``')
+        self.TAG_QUOTATION_END = self.register_tag('QuotationEnd', "''")
+
         self.RE_TAG = re_compile(
             r'([^' + re_char_delim + r'\s]+)' + re_char_delim
           + r'([^' + re_char_delim + r'\s]+)')
         self.TAG_UNK = self.register_tag('EntityUnk')
 
+        self.SHORTCUTS = []
+        self.TAG_SHORTCUT = self.CHAR_DELIM + self.CHAR_DELIM + 'Shortcut'
+
+    def add_shortcut(self, orig, subst):
+        res = ''
+        for subst_ in subst.split():
+            idx = len(self.SHORTCUTS)
+            res += '{}{}{}' \
+                       .format('' if orig else ' ', idx, self.TAG_SHORTCUT)
+            self.SHORTCUTS.append((subst_, orig))
+            orig = ''
+        return res
+
     def clear_corpus(self):
         self._corpus = OrderedDict()
 
@@ -350,7 +373,8 @@ class TextPreprocessor:
         print('Load corpus...', end=' ', file=LOG_FILE)
         LOG_FILE.flush()
         with open(path, mode='rt', encoding=encoding) as f:
-            res = self.new_pars(list(self.text_to_pars(f.read(), eop=eop)), doc_id)
+            res = self.new_pars(list(self.text_to_pars(f.read(), eop=eop)),
+                                doc_id)
         print('done.', file=LOG_FILE)
         return res
 
@@ -409,22 +433,22 @@ class TextPreprocessor:
         text = self.RE_EMOJI.sub(
             lambda x:
                 (x.group(1) + ' ' + x.group(2) + self.TAG_EMOJI + ' '
-                                                                   + x.group(3)
+                                                                  + x.group(3)
                      if x.group(2) else '')
               + (x.group(4) + x.group(5) + self.TAG_EMOJI + ' ' + x.group(6)
                      if x.group(5) else '')
               + (x.group(7) + ' ' + x.group(8) + self.TAG_EMOJI + ' '
-                                                                   + x.group(9)
+                                                                  + x.group(9)
                      if x.group(8) else '')
               + (' ' + x.group(10) + self.TAG_EMOJI + ' '
                      if x.group(10) else '')
               + (x.group(11) + ' ' + x.group(12) + self.TAG_EMOJI + ' '
                      if x.group(12) else '')
-              + (x.group(13) + x.group(14) + ' ' + x.group(15) + self.TAG_EMOJI
-                                                                          + ' '
+              + (x.group(13) + x.group(14) + ' ' + x.group(15)
+                                                        + self.TAG_EMOJI + ' '
                      if x.group(15) else '')
               + (' ' + x.group(16) + self.TAG_EMOJI + ' ' + x.group(17)
-                                                                  + x.group(18)
+                                                                 + x.group(18)
                      if x.group(16) else '')
               + (' yandex_' + x.group(19) + self.TAG_EMOJI
                      if x.group(19) else ''),
@@ -534,7 +558,7 @@ class TextPreprocessor:
 
     def _tag_date(self, text):
         def process(match):
-            res = token = match.group(0)
+            res = match.group(0)
             d, m, y = int(match.group(1)), int(match.group(2)), match.group(3)
             if len(y) == 2: y = '20' + y
             y = int(y)
@@ -561,6 +585,21 @@ class TextPreprocessor:
         )
         return text
 
+    def _tag_quotation(self, text):
+        def process(match):
+            res = match.group(0)
+            for i in range(1, 12, 3):
+                q1 = match.group(i)
+                if q1:
+                    q2, q3 = match.group(i + 1), match.group(i + 2)
+                    res = q1 + self.TAG_QUOTATION_START + ' ' \
+                        + q2 + ' ' \
+                        + q3 + self.TAG_QUOTATION_END + ' '
+                    break
+            return res
+        text = self.RE_QUOTATION.sub(process, text)
+        return text
+
     def norm_punct(self, text, islf_eos=True, istab_eos=True,
                    ignore_case=False):
         """Some heuristics to normalize Russian punctuation. Use it for chat
@@ -596,7 +635,7 @@ class TextPreprocessor:
         text = text.replace(r'""', r' " ')
         text = re_sub(r',,?', r' , ', text)
         text = re_sub(r'№№?', r' № ', text)
-        # апостроф в начале или в конце слова - кавычки
+        # апостроф в начале или в конце строки - кавычки
         text = re_sub(r"^'|'$", '"', text)
         # если несколько символов ., ?, !, подряд, то если среди них есть
         # ?, то меняем всё на него, если есть !, то на него, иначе ставим
@@ -621,44 +660,94 @@ class TextPreprocessor:
             (r'[ЁА-Я](?:[ёа-я]+-[ЁА-Я])?[ёа-я]+', r'[ЁА-Я]\.')
         ]:
             # инициалы в начале:
-            text = re_sub(r'\b({})({})? ?({})\b'
-                              .format(re_init, re_init, re_lname),
+            text = re_sub(r'\b({0})({0})? ?({1})\b'
+                              .format(re_init, re_lname),
                           r' \g<1> \g<2> \g<3> ', text, flags=flags)
             # инициалы в конце:
-            text = re_sub(r'\b({}) ({})({})?\b'
-                              .format(re_init, re_init, re_lname),
+            text = re_sub(r'\b({1}) ({0})({0})?\b'
+                              .format(re_init, re_lname),
                           r' \g<1> \g<2> \g<3> ', text, flags=flags)
 
-        # period just before a word
-        text = re_sub(r'(^|\W)\.(\w)', r'\g<1>. \g<2>', text)
-
         # --- end of sentence w/o space after period ---
         def process(match):
             a, b = match.groups()
-            return a + '. ' + b if wform_isknown(b) or wform_isknown(a) else \
-                   match.group(0)
+            return a + '. ' + b if b.lower() not in [
+                'com', 'org', 'edu', 'net', 'info',
+                'de', 'cn', 'uk', 'ru', 'su', 'us', 'jp',
+                'бг', 'бел', 'рф', 'срб', 'укр'
+            ] and (wform_isknown(a) or wform_isknown(b)) else match.group(0)
         text = re_sub(r'(\w+)\.(\w+)', process, text)
         text = re_sub(r'(\w+)\.(\w+)', process, text)  # sic!
 
+        # period just before a word
+        text = re_sub(r'(^|\W)\.(\w)', r'\g<1>. \g<2>', text)
+
+        # period before of quotation:
+        text = re_sub(r'(\w+)\.\s*(["`«„]\s*\b)', r'\g<1> . \g<2>', text)
+
+        # known bugs of russian nltk punkt:
+        text = re_sub(r'\b(я|театр|нас|прав)\.', r'\g<1> .', text)
+
         # --- known shortcuts ---
+        '''
         re_0 = r'\b'
         re_1 = r'\b\.?\s*([ЁА-Я])?' # конец слова; дальше м.б. точка и/или
-                                    # заглавная буква через пробелы или без них
+                                    # заглавная буква через пробелы или без
+                                    # них
         re_2 = r'\s*\.?\s*'
         re_3 = r'\b\s*\.?\s*'       # конец слова, после которого возможны
                                     # пробелы и/или точка
         re_4 = r'\s*'
         re_5 = r'\s+'
-        for a, b in [('{0}[иИ]{4}т{2}д{1}', 'и так далее'),
-                     ('{0}[иИ]{4}т{2}п{1}', 'и тому подобное'),
-                     ('{0}[мМ]{3}б{1}',     'может быть'),
-                     ('{0}[тТ]{3}е{1}',     'то есть'),
-                     ('{0}[тТ]{2}к{1}',     'так как')]:
+        #TODO: capitalization
+        for a, b in [(r'{0}[иИ]{4}т{2}д{1}', r'и так далее'),
+                     (r'{0}[иИ]{4}т{2}п{1}', r'и тому подобное'),
+                     (r'{0}[мМ]{3}б{1}',     r'может быть'),
+                     (r'{0}[тТ]{3}е{1}',     r'то есть'),
+                     (r'{0}[тТ]{2}к{1}',     r'так как')]:
             text = re_sub(a.format(re_0, re_1, re_2, re_3, re_4, re_5),
-                          # если после сокращения идёт слово с заглавной буквы,
-                          # то ставим перед ним точку
-                          lambda x: ' {} {}'.format(b, '. ' + x.group(1)
-                                                        if x.group(1) else ''),
+                          # если после сокращения идёт слово
+                          # с заглавной буквы, то ставим перед ним точку
+                          lambda x: ' {} {}'
+                                        .format(b, ('. ' + x.group(1))
+                                                       if x.group(1) else ''),
+                          text)
+        for a, b in [(r'{0}г-ж([аеиу]|ой){0}', r'госпож\g<1>'),
+                     (r'{0}г-н([аеу]|ом)?{0}', r'господин\g<1>')]:
+            text = re_sub(a.format(re_0), ' {} '.format(b), text)
+        '''
+        re_0 = r'\b'
+        re_1 = r'\s*([ЁА-Я])?'  # заглавная буква через пробелы или без них
+        re_2 = r'\b\s*\.?'      # конец слова, после которого возможны пробелы
+                                # и/или точка
+        re_3 = re_2 + r'\s*'    # то же, что и re_2, но в конце ещё может быть
+                                # пробел
+        re_4 = r'\s*'
+        re_5 = r'\s+'
+        #TODO: capitalization
+        for a, b in [(r'({0}[иИ]{4}т{2}д{2}){1}',  r'и так далее'),
+                     (r'({0}[иИ]{4}т{2}п{2}){1}',  r'и тому подобное'),
+                     (r'({0}[мМ]{3}б{2}){1}',      r'может быть'),
+                     (r'({0}[тТ]{3}е{2}){1}',      r'то есть'),
+                     (r'({0}[тТ]{3}к{2}){1}',      r'так как')]:
+            text = re_sub(a.format(re_0, re_1, re_2, re_3, re_4, re_5),
+                          # если после сокращения идёт слово
+                          # с заглавной буквы, то ставим перед ним точку
+                          lambda x: ' {} {}'
+                                        .format(self.add_shortcut(x.group(1),
+                                                                  b),
+                                                ('. ' + x.group(2))
+                                                    if x.group(2) else ''),
+                          text)
+        for a, b in [(r'({0}г-ж([аеиу]|ой){0})', r'госпож'),
+                     (r'({0}г-н([аеу]|ом)?{0})', r'господин')]:
+            text = re_sub(a.format(re_0),
+                          lambda x: ' {} '.format(
+                              self.add_shortcut(x.group(1),
+                                                b + (x.group(2)
+                                                         if x.group(2) else
+                                                     ''))
+                          ),
                           text)
 
         # === HYPHENS ===
@@ -666,12 +755,23 @@ class TextPreprocessor:
 
         # --- searching dashes between hyphens ---
         def process(match):
-            token = match.group(0)
+            # если один из токенов - наш тэг, то ничего не меняем
+            if self.CHAR_DELIM in [match.group(1), match.group(3)]:
+                return match.group(0)
+
+            token = match.group(2)
             # сохраняем разделители
             hyphens = re_findall('\W+', token)
 
             res = ''
             words = token.replace(' ', '').split('-')
+            test_word = '{}-{}'.format(words[0], words[1])
+            if len(words) == 2 and (
+                wform_isknown(test_word) or (
+                    words[0].isdecimal() and words[1].isalpha()
+                )
+            ):
+                return '{}-{}'.format(words[0], words[1])
             # поиск:                   -i-
             #                 [xxx....] 0
             #                 [.xxx...] 1
@@ -731,29 +831,29 @@ class TextPreprocessor:
             while True:
                 has1more = i > 0
                 if i >= 2:
-                    for word in [words[i-2] + '-' + words[i-1],
-                                 words[i-2] + ''  + words[i-1]]:
+                    for word in [words[i - 2] + '-' + words[i - 1],
+                                 words[i - 2] + ''  + words[i - 1]]:
                         if wform_isknown(word):
                             res += ' ' + word
                             has1more = False
                             maybehyphen = 0
                             break
                     else:
-                        add_word(i-2)
+                        add_word(i - 2)
                 if i >= len_words:
                     if has1more:
-                        add_word(i-1)
+                        add_word(i - 1)
                     break
                 if i <= last_3:
                     for word in [
-                        words[i] + '-' + words[i+1] + '-' + words[i+2],
-                        words[i] + ''  + words[i+1] + '-' + words[i+2],
-                        words[i] + '-' + words[i+1] +  '' + words[i+2],
-                        words[i] +  '' + words[i+1] +  '' + words[i+2]
+                        words[i] + '-' + words[i + 1] + '-' + words[i + 2],
+                        words[i] + ''  + words[i + 1] + '-' + words[i + 2],
+                        words[i] + '-' + words[i + 1] +  '' + words[i + 2],
+                        words[i] +  '' + words[i + 1] +  '' + words[i + 2]
                     ]:
                         if wform_isknown(word):
                             if has1more:
-                                add_word(i-1)
+                                add_word(i - 1)
                             res += ' ' + word
                             words = words[i + 3:]
                             len_words = len(words)
@@ -770,78 +870,62 @@ class TextPreprocessor:
             return res + ' '
 
         # находим все слова c дефисами; с одной стороны от дефиса м.б. пробел
-        text = re_sub(r'\w+(?:(?:-| -|- )\w+)+', process, text)
+        text = re_sub(r'(\{})?(\w+(?:(?:-| -|- )\w+)(\{})?)+'
+                           .format(self.CHAR_DELIM, self.CHAR_DELIM),
+                      process, text)
+
+        # дефис в начале русского слова = тире
+        text = re_sub(r'(^|[^0-9ЁА-Яёа-я])-([ЁА-Яёа-я])', '\g<1>- \g<2>',
+                      text)
+        # дефис после знака препинания = тире
+        text = re_sub(r'([.!?])-(\s|$)', '\g<1> -\g<2>', text)
 
         return text
 
-    @staticmethod
-    def sent_tokenize(text, kill_empty=True):
+    def sent_tokenize(self, text, kill_empty=True):
         """Return sentence-tokenized copy of a *text*
 
         :rtype: list
         """
+        text = text.replace('«', '``').replace('„', '``') \
+                   .replace('»', "''").replace('“', "''")
+
         sents = nltk_sent_tokenize(text, language='russian')
-        """
-        sents = []
-        prev_sent = None
-        for sent in sents_:
-            sent = sent.strip()
-            with open('111', 'at', encoding='utf-8') as f:
-                print(sent, file=f)
-            if sents:
-                if sent[0].isalpha():
-                    # если первая буква предложения - строчная, то приклеиваем
-                    # предложение к предыдущему
-                    if not sent[0].istitle():
-                        prev_sent += ' ' + sent
-                        sents[-1] = prev_sent
-                        continue
-                    #elif prev_sent[-1] == '"' and prev_sent.count('"') % 2:
-                    # если в конце предыдущего предложения кавычка, но
-                    # начинается оно не с кавычки, то приклеиваем к нему
-                    # следующее
-                    elif prev_sent[-1] == '"' and prev_sent[0] != '"':
-                        sents[-1] = prev_sent[:-1]
-                        sent = '"' + sent
-                    else:
-                        # если в конце предыдущего предложения инициал, а новое
-                        # начинается с инициала, то склеиваем.
-                        # если же в первом слове нового предложения только
-                        # буквы, а в конце предыдущего инициалы, перед которыми
-                        # слово не с большой буквы (либо с большой, но оно
-                        # первое в предложении), то тоже склеиваем.
-                        #
-                        # инициал в конце предыдущей: "X." или "Xx."
-                        if re_search(r'\b' + self._CARE + r'?\b\s*\.$', prev_sent):
-                            # инициал в начале следующей
-                            # либо
-                            # в конце инициалы а перед ними не фамилия,
-                            # а в начале следующей - фамилия
-                            if re_match(r'\b' + self._CARE + r'?\b\s*\.', sent) \
-                            or (re_search(r'''(?x)
-                                (?: ^                 # либо начало строки
-                                  | (?: ^ |'''        # либо начало строки или
-                                      + self._NOCASP  #     не заглавная буква, за которыми идут
-                              + r''') \s* [^\s\w]\s*  #     не буква и не пробел, но потом м.б. пробелы
-                                  | \b                              # либо начало слова - такое, что:
-                                      (?: ^''' + self._CAPS + r'''  #     первая буква м.б. заглавной только в начале строки
-                                        |''' + self._NOCA + r'''    #     иначе первая буква - строчная
-                                      ) \S*? \b \s+                 # дальше м. идти любая непробельная дичь
-                                )                                   #      до конца слова, а потом хотя бы один пробел
-                                (?: \b''' + self._CARE + r'''?\b\s*\.\s* )*  # м.б. инициалы
-                                \b''' + self._CARE + r'''?\b\s*\.$''',       # в конце инициал
-                            prev_sent)
-                            and re_match(r'\b' + self._CARE + r'+\b\s*(?:[^.]|\.$)',
-                                         sent)):
-                                prev_sent += ' ' + sent
-                                sents[-1] = prev_sent
-                                continue
-            prev_sent = sent
-            sents.append(sent)
-        """
+
+        re_ellipsis = re_compile(r'(\.\.\.)\s+([A-ZЁЯ-Я])')
+        def parse_el(sent):
+            sents = []
+            ellipsis = self.CHAR_DELIM + 'ellipsis' + self.CHAR_DELIM
+            len_ellipsis = len(ellipsis)
+            sent = re_ellipsis.sub('\g<1>{}\g<2>'.format(ellipsis), sent)
+            i = 0
+            while True:
+                i = sent.find(ellipsis, i)
+                if i == -1:
+                     break
+                sents.append(sent[:i])
+                sent = sent[i + len_ellipsis:]
+            if sent:
+                sents.append(sent)
+            return sents
+
+        sents_ = []
+        re_quot = re_compile(r'\d+' + '\\' + self.TAG_QUOTATION_END)
+        for sent in sents:
+            match = re_quot.match(sent)
+            if sents_ and match:
+                quot = match.group(0)
+                sents_[-1] += ' ' + quot
+                sent = sent[len(quot):]
+
+            for s in parse_el(sent):
+                sents_.append(s)
+
         if kill_empty:
-            sents = filter(lambda x: re_search('(?i)[0-9a-zёа-я]', x), sents)
-        return sents
+            sents_ = list(filter(lambda x: re_search('(?i)[0-9a-zёа-я]', x),
+                                 sents_))
+
+        return sents_
 
     @staticmethod
     def word_tokenize(text):
@@ -852,6 +936,19 @@ class TextPreprocessor:
         # NB: "" -> ``''
         tokens = nltk_word_tokenize(text, language='russian',
                                     preserve_line=True)
+        try:
+            idx = tokens.index("'")
+        except ValueError:
+            try:
+                idx = tokens.index("’")
+            except ValueError:
+                idx = -1
+        if idx > 0 and tokens[idx - 1].isalpha() \
+                   and tokens[idx + 1].isalpha() \
+                   and tokens[idx + 1].istitle():
+            tokens = tokens[:idx - 1] \
+                   + [tokens[idx - 1] + tokens[idx] + tokens[idx + 1]] \
+                   + tokens[idx + 2:]
         return tokens
 
     @staticmethod
@@ -1067,13 +1164,14 @@ class TextPreprocessor:
                    default_tagger(text) if tagger else \
                    text
 
+        tag_quotation = True
         for tagger, default_tagger in zip(
             [pre_tag, tag_emoji, tag_xml, tag_email,
              tag_uri, tag_phone, tag_date, tag_hashtag,
-             tag_nametag, post_tag],
+             tag_nametag, tag_quotation, post_tag],
             [lambda x: x, self._tag_emoji, self._tag_xml, self._tag_email,
              self._tag_uri, self._tag_phone, self._tag_date, self._tag_hashtag,
-             self._tag_nametag, lambda x: x]
+             self._tag_nametag, self._tag_quotation, lambda x: x]
          ):
             text = run_tagger(tagger, default_tagger)
 
@@ -1099,20 +1197,25 @@ class TextPreprocessor:
             space_before = False
             for i, token in enumerate(tokens):
                 wform = token['FORM']
-                delim_pos = wform.rfind(self.CHAR_DELIM)
+                delim_pos = wform.find(self.CHAR_DELIM)
                 misc = token['MISC']
                 if delim_pos >= 0:
                     idx = int(wform[:delim_pos])
                     tag = wform[delim_pos:]
-                    mask = self.TAG_MASKS[tag]
-                    tag = tag[1:]
-                    orig = tags[tag][idx]
-                    token = tokens[i]
-                    token['FORM'] = mask
-                    misc[tag] = orig
-                    if space_before:
-                        text += ' '
-                    text += orig
+                    if tag == self.TAG_SHORTCUT:
+                        subst, orig = self.SHORTCUTS[idx]
+                        token['FORM'] = subst
+                        misc[self.TAG_SHORTCUT[2:]] = orig
+                    else:
+                        mask = self.TAG_MASKS[tag]
+                        tag = tag[1:]
+                        orig = tags[tag][idx]
+                        #token = tokens[i]
+                        token['FORM'] = mask
+                        misc[tag] = orig
+                        if space_before:
+                            text += ' '
+                        text += orig
                 elif wform in ['``', '(', '«']:
                     misc['SpaceAfter'] = 'No'
                     if space_before:
